@@ -2,10 +2,13 @@ from __future__ import print_function
 
 from datetime import datetime
 from warnings import warn
+from socket import AF_INET, SOCK_STREAM, socket
 import argparse
 import os
+from pathlib import Path
 import traceback
-import stat
+import shlex
+import subprocess
 import sys
 
 from . import settings
@@ -17,14 +20,6 @@ from .labelprinterapiclient import get_labelprinter
 
 S_UNSET = '(unset)'
 
-try:
-    raw_input  # py2
-except NameError:
-    def raw_input(msg=None):  # py3
-        if msg is not None:
-            print(msg, end='')
-        return input()
-
 
 class Inventory:
     def __init__(self, inventory_rest_client, hdd_id):
@@ -108,7 +103,7 @@ class DeviceManager(object):
         health = (
             'OK' if self._hdd_dock.hwdata.wear_health_percent >= 100
             else '{}%'.format(self._hdd_dock.hwdata.wear_health_percent))
-        health_status = f'{health} - {title}'
+        health_status = f'health {health} - {title}'
         self.set_health(health_status)
 
     def set_health(self, health_status):
@@ -155,17 +150,24 @@ class DeviceManager(object):
         # XXX: better dynamic value for sample count based on disk size?
         sample_count = settings.POST_WIPE_SAMPLE_COUNT
         if self._hdd_dock.is_ssd():
-            sample_count *= 20  # SSDs/NVMes are fast(er)
+            sample_count *= 10  # SSDs/NVMes are fast(er)
         sample_size = settings.POST_WIPE_SAMPLE_SIZE
 
         print('Attempting erase method {!r} on BLKDEV {!r}'.format(
             erase_method.name, self._hdd_dock.devname))
 
-        # Running pre-sample! This is not needed, but nice during dev.
-        if self._find_non_zero_blocks(sample_count, sample_size):
-            print('DEBUG: Found non-zero samples before wipe (expected)')
-        else:
+        # Running pre-sample.  This is not needed, but nice during dev.
+        print(
+            'DEBUG: Sampling block device to check if empty... '
+            '(sample_count: {}, sample_size {})'.format(
+                sample_count, sample_size))
+        sampler = DiskEmptySampler(
+            self._hdd_dock.devname, sample_count, sample_size)
+        sampler.sample()
+        if sampler.is_zero():
             print('DEBUG: No non-zero sample found. Disk already zeroed?')
+        else:
+            print('DEBUG: Found non-zero samples before wipe (expected)')
 
         print('Start secure wiping BLKDEV')
         success, error = erase_method()
@@ -181,8 +183,16 @@ class DeviceManager(object):
         print('Block device securely wiped')
 
         # Running post-sample. Should be all zero.
-        if self._find_non_zero_blocks(sample_count, sample_size):
-            print('ERROR: Found non-empty sample!')
+        print('DEBUG: Sampling block device afterwards...')
+        sampler.sample()
+        if sampler.is_zero():
+            print('OK: All samples are empty')
+        elif sampler.is_different():
+            print(
+                '(probably) OK: All samples are different '
+                '(old school crypto disk?)')
+        else:
+            print('ERROR: Found not-wiped sample!')
             print('Please check {!r} manually!'.format(self._hdd_dock.devname))
             self._set_status(
                 'BLKDEV_SECURE_WIPED_ERROR', (
@@ -192,11 +202,11 @@ class DeviceManager(object):
             self._set_health('ERASE ERROR')
             return
 
-        print('OK: All samples are empty')
+        # All good!
         self._set_status(
             'BLKDEV_SECURE_WIPED', (
                 'Block device secure wiped using {!r} at {} and '
-                'checked {} {} --{}'
+                'checked {}x{} --{}'
                 .format(
                     erase_method.name, self._location, sample_count,
                     sample_size, self._author)))
@@ -206,10 +216,6 @@ class DeviceManager(object):
             self._set_health('SECURE ERASED')
 
     def _find_non_zero_blocks(self, sample_count, sample_size):
-        print(
-            'DEBUG: Sampling block device to check if empty... '
-            '(sample_count: {}, sample_size {})'.format(
-                sample_count, sample_size))
 
         disk_empty_sampler = DiskEmptySampler(
             self._hdd_dock.devname, sample_count, sample_size)
@@ -230,20 +236,20 @@ def register_disk(**kwargs):
 
     print('')
     print('Disk is not registered yet, please specify the following fields:')
-    owner = raw_input('Owner [OSSO]: ')
+    owner = input('Owner [OSSO]: ')
 
     # Default to OSSO
     if owner in (None, ''):
         owner = 'OSSO'
 
-    bay = raw_input('bay ['']: ')
+    bay = input('bay ['']: ')
 
     if bay in (None, ''):
         bay = ''
 
     erase = None
     while erase not in ('y', 'n', 'Y', 'N', ''):
-        erase = raw_input('Quick erase? (y/N): ')
+        erase = input('Quick erase? (y/N): ')
 
     result = inventory_rest_client.add_hdd(hdd_dock, bay)
     tag_uid = result['tag_uid']
@@ -270,10 +276,10 @@ def register_disk(**kwargs):
         labelprinter_rest_client.print_hdd_label(
             tag_uid, hdd_dock.get_serial_number(), owner)
 
-    raw_input('Registration complete, press enter')
+    input('Registration complete, press enter')
 
 
-def registered_disk_actions(**kwargs):
+def registered_disk_actions(action=None, **kwargs):
     # Python2-compatible forced kwargs. Temporary fix until we clean up this
     # code.
     hdd_dock = kwargs.pop('hdd_dock')
@@ -317,15 +323,25 @@ def registered_disk_actions(**kwargs):
     manager = DeviceManager(
         hdd_dock, hdd_id, inventory_rest_client, author, location)
     manager.show_summary(_build_print_data())
+
+    if action == 'secure-erase':
+        if 'SECURE ERASED' in inventory.current_health_status:
+            print(
+                'Skipping automatic secure erase, health says it\'s been done')
+        else:
+            print('Starting automatic secure erase')
+            manager.best_erase()
+            inventory.flush()
+
     manager.show_commands()
 
     while True:
-        action = raw_input('\nAction: ').upper()
+        action = input('\nAction: ').upper()
 
         if action == '1':
             owner = ''
             while len(owner) == 0:
-                owner = raw_input('New owner: ')
+                owner = input('New owner: ')
             inventory_rest_client.add_owner(  # XXX: not email-of-owner
                 hdd_id, name=owner, email=author)
             inventory.flush()
@@ -348,47 +364,20 @@ def registered_disk_actions(**kwargs):
         elif action == '5':
             bay = ''
             while len(bay) == 0:
-                bay = raw_input('New server bay: ')
+                bay = input('New server bay: ')
             inventory_rest_client.change_bay(hdd_id, bay)
             inventory.flush()
             print('Server bay changed to: {}'.format(inventory.bay))
 
         elif action == '6':
             health_status = (
-                raw_input('Manual health status [ONLINE]: ') or 'ONLINE')
+                input('Manual health status [ONLINE]: ') or 'ONLINE')
             manager.set_health_with_wear(health_status)
             inventory.flush()
 
         elif action == '7':
             # Print label
-            today = datetime.now().strftime('%Y-%m-%d')
-            # XXX: don't use dynamic_data please :(
-            total_bytes_written = [
-                x for x in dynamic_data
-                if 'Total bytes written' in x[0]][0]
-            total_bytes_read = [
-                x for x in dynamic_data
-                if 'Total bytes read' in x[0]][0]
-
-            lines = [
-                'Health status : {}'.format(
-                    inventory.current_health_status.upper()),
-                '',
-                'Serial : {}'.format(hdd_dock.get_serial_number()),
-                'Total bytes written/read : {}/{}'.format(
-                    total_bytes_written[1], total_bytes_read[1]),
-                'Power on hours : {} hours'.format(
-                    hdd_dock.hwdata.power_on_hours or S_UNSET),
-                f'--{author} @ {today}',
-            ]
-            try:
-                labelprinter_rest_client = get_labelprinter()
-                labelprinter_rest_client.print_generic_label(lines)
-            except Exception:
-                print('Would have printed:')
-                print('\n'.join('- {}'.format(line) for line in lines))
-                print()
-                raise
+            print_health_label(hdd_dock, inventory, author, dynamic_data)
 
         elif action == '8':
             manager.second_best_erase()
@@ -416,7 +405,7 @@ def registered_disk_actions(**kwargs):
             print('Notice: {!r} action unknown'.format(action))
 
     hdd_dock.shutdown_disk()
-    exit(0)
+    sys.exit(0)
 
 
 def human_readable_bytes(byte_count):
@@ -455,6 +444,79 @@ def do_print(data, hdd_dock):
             print(fmt.format(key, item))
 
 
+def print_health_label(hdd_dock, inventory, author, dynamic_data):
+    """
+    Print health status on a label. See also print_health_label_from_db.
+    Both would like some refactoring.
+    """
+    today = datetime.now().strftime('%Y-%m-%d')
+    # XXX: don't use dynamic_data please :(
+    total_bytes_written = [
+        x for x in dynamic_data
+        if 'Total bytes written' in x[0]][0]
+    total_bytes_read = [
+        x for x in dynamic_data
+        if 'Total bytes read' in x[0]][0]
+
+    lines = [
+        'Health status : {}'.format(
+            inventory.current_health_status.upper()),
+        '',
+        'Serial : {}'.format(hdd_dock.get_serial_number()),
+        'Total bytes written/read : {}/{}'.format(
+            total_bytes_written[1], total_bytes_read[1]),
+        'Power on hours : {} hours'.format(
+            hdd_dock.hwdata.power_on_hours or S_UNSET),
+        f'--{author} @ {today}',
+    ]
+    try:
+        labelprinter_rest_client = get_labelprinter()
+        labelprinter_rest_client.print_generic_label(lines)
+    except Exception:
+        print('Would have printed:')
+        print('\n'.join('- {}'.format(line) for line in lines))
+        print()
+        raise
+
+
+def print_health_label_from_db(hdd_data):
+    """
+    A shabby version of print_health_label() using all info we have
+    available in the DB.
+    """
+    print(
+        'WARNING: total bytes read/written and total power hours '
+        'not known when printing; you should have printed a health label '
+        'in "dev" mode', file=sys.stderr)
+    print('... but printing a partial label for you anyway.', file=sys.stderr)
+
+    health = hdd_data['current_health']['status'].upper()
+    author_of_health = hdd_data['current_health']['extra_info'].strip()
+    if '--' in author_of_health:
+        author_of_health = author_of_health.rsplit('--', 1)[-1]
+    else:
+        author_of_health = 'unknown'
+    serial = hdd_data['serial_number']
+    date_of_health = hdd_data['current_health']['timestamp'].split('T')[0]
+
+    lines = [
+        f'Health status : {health} @ {date_of_health}',
+        '',
+        f'Serial : {serial}',
+        'Total bytes written/read : (not known in db)',
+        'Power on hours : (not known in db)',
+        f'--{author_of_health} @ {date_of_health}',
+    ]
+    try:
+        labelprinter_rest_client = get_labelprinter()
+        labelprinter_rest_client.print_generic_label(lines)
+    except Exception:
+        print('Would have printed:')
+        print('\n'.join('- {}'.format(line) for line in lines))
+        print()
+        raise
+
+
 def build_hdd_info(hdd_dock):
     hwdata = hdd_dock.hwdata
 
@@ -511,16 +573,22 @@ def build_hdd_info(hdd_dock):
     return static_data, dynamic_data
 
 
-def main_menu(devname):
+def get_author_and_location():
     author = os.environ.get('EMAIL', '')
     if '@' not in author:
         print('ERROR: Please set the EMAIL envvar to specify who you are!')
-        exit(1)
+        sys.exit(1)
     location = os.environ.get('LOCATION', 'remote')
     if not location:
         print('ERROR: Please (un)set the LOCATION envvar to specify where!')
-        exit(1)
+        sys.exit(1)
+    return author, location
 
+
+def dev_menu(devname, action):
+    if os.getuid() != 0:
+        warn('Expected UID 0 (root) for access')
+    author, location = get_author_and_location()
     hdd_dock = BaseStorageDevice.from_devname(devname)
     sys.stdout.write('\x1b]2;DOCKTOOL DISK BAY: {}\x07'.format(
         hdd_dock.docktool_bay_nr))
@@ -560,22 +628,217 @@ def main_menu(devname):
             hdd_dock.get_device_model(),
             hdd_dock.get_serial_number())
         assert hdd_id, 'HDD id not set after registration?'
+
+    registered_disk_actions(
+        hdd_dock=hdd_dock, hdd_id=hdd_id,
+        inventory_rest_client=inventory_rest_client,
+        static_data=static_data, dynamic_data=dynamic_data,
+        author=author, location=location, action=action)
+
+
+def db_menu(serial):
+    inventory_rest_client = InventoryRESTClient(
+        settings.DASHBOARD_BASE_URL)
+
+    # HDD is an asset so asset_id = hdd_id
+    hdd_id = inventory_rest_client.get_hdd_id(None, serial)  # serial?
+    if hdd_id is not None:
+        hdd = inventory_rest_client.get_hdd(hdd_id)[0]
     else:
-        registered_disk_actions(
-            hdd_dock=hdd_dock, hdd_id=hdd_id,
-            inventory_rest_client=inventory_rest_client,
-            static_data=static_data, dynamic_data=dynamic_data,
-            author=author, location=location)
+        hdd = inventory_rest_client.get_hdd(serial)[0]  # hdd_id as "serial"
+    if hdd is None:
+        raise ValueError('serial/asset not found in remote DB')
+
+    # Ok, we have an ID. Get info from remote?
+    do_print_health_label = None
+    while do_print_health_label not in ('y', 'n', ''):
+        do_print_health_label = input('Print health label? (y/N): ').lower()
+
+    if do_print_health_label == 'y':
+        print_health_label_from_db(hdd)
+        print('done.')
+    else:
+        from pprint import pprint
+        pprint(hdd)
+
+
+class AutoServer:
+    def __init__(self, action, gui, ignore_devices):
+        self.action = action
+        self.gui = gui
+        self.ignore_devices = [i.resolve() for i in ignore_devices]
+        # Make sure the environment is valid to start osso-docktool.
+        author, location = get_author_and_location()
+        if not gui:
+            if not os.environ.get('TMUX'):
+                os.execvp('tmux', [
+                    'tmux', 'new', '-s', 'docktool', '-n', 'main',
+                    f'{self.docktool_exec} auto {action}'])
+            else:
+                # Add the environment to the tmux session.
+                subprocess.check_call(
+                    ['tmux', 'set', 'remain-on-exit', 'on'])
+                subprocess.check_call(
+                    ['tmux', 'setenv', 'EMAIL', author])
+                subprocess.check_call(
+                    ['tmux', 'setenv', 'LOCATION', location])
+
+    def start(self):
+        try:
+            self.listen()
+        except KeyboardInterrupt:
+            raise SystemExit()
+
+    def listen(self):
+        with socket(AF_INET, SOCK_STREAM) as sock:
+            sock.bind(('127.0.0.1', 1451))
+            sock.listen(4)
+            sock.setblocking(True)
+            if self.ignore_devices:
+                print(
+                    'Ignoring devices',
+                    ','.join([str(i) for i in self.ignore_devices]))
+            print('Waiting for new disk notifications...')
+            while True:
+                connection, address = sock.accept()
+                pid = os.fork()
+                if pid == 0:
+                    self.handle_client(connection)
+                else:
+                    os.waitid(os.P_PID, pid, os.WEXITED)
+
+    def handle_client(self, connection):
+        client_data = b''
+        with connection:
+            while True:
+                data = connection.recv(1024)
+                if not data:
+                    break
+                client_data += data
+
+        if client_data:
+            device = Path(client_data.decode())
+            if device.is_block_device():
+                if self.is_ignored(device):
+                    print(f'Ignoring notification for {device}')
+                else:
+                    self.start_docktool(device)
+            else:
+                print(f'Path {device} is not a block device')
+
+    @property
+    def docktool_exec(self):
+        return Path(sys.argv[0]).resolve()
+
+    def is_ignored(self, device):
+        # Check the resolved path against the ignore list.
+        return bool(device.resolve() in self.ignore_devices)
+
+    def start_docktool(self, device):
+        if self.gui:
+            self.start_gui(device)
+        else:
+            self.start_cli(device)
+
+    def start_gui(self, device):
+        print(f'Starting GUI terminal for {device}')
+        # Double fork to detach terminal from pid group.
+        if os.fork() == 0:
+            args = [
+                f'osso-docktool-terminal [{device}]', '-e',
+                'sudo', '-E', self.docktool_exec, 'dev', device, '--wait']
+            if self.action != 'start':
+                args += ['--action', self.action]
+            os.setsid()
+            os.execvp('x-terminal-emulator', args)
+        else:
+            os._exit(0)
+
+    def start_cli(self, device):
+        print(f'Creating new tmux window for {device}')
+        if not os.environ.get('TMUX'):
+            print('Error: TMUX environment variable is empty', file=sys.stderr)
+        else:
+            args = ['tmux', 'new-window', '-n', device]
+            cmd = (
+                f'sudo -E {self.docktool_exec} dev {shlex.quote(str(device))} '
+                '--wait')
+            if self.action != 'start':
+                cmd += f' --action {self.action}'
+            subprocess.check_call(args + [cmd])
+        os._exit(0)
+
+
+def auto_notify(device):
+    with socket(AF_INET, SOCK_STREAM) as sock:
+        try:
+            sock.connect(('127.0.0.1', 1451))
+            sock.sendall(str(device).encode())
+        except ConnectionError as e:
+            print(f'Error notifying auto process: {e}', file=sys.stderr)
 
 
 class DocktoolArgumentParser(argparse.ArgumentParser):
     def __init__(self):
         super().__init__(
+            prog='osso-docktool',
             description='Docktool for processing disks')  # exit_on_error=False
-        self.add_argument(
+        self.message2 = None
+
+        self.subparsers = self.add_subparsers(
+            title='subcommands', dest='command', required=True,
+            help='subcommand help',
+            parser_class=argparse.ArgumentParser)
+
+        devparser = self.subparsers.add_parser(
+            'dev',
+            help='(implied!) process hardware device (original behaviour)')
+        devparser.add_argument(
             'device', metavar='DISK', type=self.block_device,
             help='Device to use for example: /dev/sda')
-        self.message2 = None
+        devparser.add_argument(
+            '-a', '--action', choices=['secure-erase'],
+            help='Automatically execute the action.')
+        devparser.add_argument(
+            '-w', '--wait', action='store_true',
+            help='Wait for user confirmation before exit.')
+
+        dbparser = self.subparsers.add_parser(
+            'db', help='process devices device info without hardware access')
+        dbparser.add_argument(
+            'serial', metavar='SERIAL',
+            help='Serial number (or GoCollect asset UID')
+
+        autoparser = self.subparsers.add_parser(
+            'auto',
+            help='Automatically perform action on newly inserted disks')
+        autoparser.add_argument(
+            '-g', '--gui', action='store_true',
+            help='Start docktool processes in a GUI instead of tmux session. '
+                 'Use `update-alternatives --config x-terminal-emulator` to '
+                 'configure the terminal.')
+        autoparser.add_argument(
+            '-i', '--ignore', action='append', metavar='DISK',
+            default=[self.block_device(i) for i in settings.IGNORE_DEVICES],
+            type=self.block_device,
+            help='Ignore the device for auto commands.')
+        autoparser.add_argument(
+            'action', choices=['start', 'secure-erase'], default='start',
+            help='Start docktool or secure erase newly inserted disks')
+
+        notifyparser = self.subparsers.add_parser(
+            'auto-notify',
+            help='Notify the auto process of the device')
+        notifyparser.add_argument(
+            'device', metavar='DISK', type=self.block_device,
+            help='Device to use for example: /dev/sda')
+
+    def parse_args(self):
+        if len(sys.argv) == 2 and not (
+                sys.argv[1].startswith('-')
+                or sys.argv[1] in self.subparsers.choices):
+            sys.argv[1:1] = ['dev']  # imply "dev"
+        return super().parse_args()
 
     def exit(self, status=0, message=None):
         if self.message2:
@@ -600,21 +863,18 @@ class DocktoolArgumentParser(argparse.ArgumentParser):
         super().exit(status=0, message=message)
 
     def block_device(self, name):
-        try:
-            st = os.stat(name)
-        except FileNotFoundError:
-            if '/' not in name:
-                ret = self.block_device('/dev/{}'.format(name))
-                warn('Prepended /dev/ to the block device name')
-                return ret
-            self.message2 = '{!r}: not found'.format(name)
-            raise argparse.ArgumentError('{}: not found'.format(name))
-        if not stat.S_ISBLK(st.st_mode):
-            self.message2 = '{!r}: not a block device'.format(name)
-            raise argparse.ArgumentError('{}: not a block device'.format(name))
-        if os.getuid() != 0:
-            warn('Expected UID 0 (root) for access')
-        return name
+        device = Path(name)
+        if not device.is_absolute():
+            ret = self.block_device('/dev' / device)
+            warn('Prepended /dev/ to the block device name')
+            return ret
+        if not device.exists():
+            self.message2 = f'{device!r}: not found'
+            raise argparse.ArgumentError(f'{device}: not found')
+        if not device.is_block_device():
+            self.message2 = f'{device!r}: not a block device'
+            raise argparse.ArgumentError(f'{device}: not a block device')
+        return device
 
 
 def main():
@@ -622,13 +882,26 @@ def main():
     args = parser.parse_args()
 
     try:
-        main_menu(args.device)
+        if args.command == 'dev':
+            dev_menu(args.device, args.action)
+        elif args.command == 'serial':
+            db_menu(args.serial)
+        elif args.command == 'auto':
+            AutoServer(args.action, args.gui, args.ignore).start()
+        elif args.command == 'auto-notify':
+            auto_notify(args.device)
+    except KeyboardInterrupt:
+        sys.exit(0)
+    except SystemExit:
+        if getattr(args, 'wait', False):
+            input('Press enter to exit')
     except Exception as e:
         print('ERROR: {}\n'.format(e), file=sys.stderr)
         print()
         print(traceback.format_exc(), file=sys.stderr)
-        raw_input('Error found, please inform developer. Press enter')
-        raise e
+        input('Error found, please inform developer. Press enter')
+        # #raise e  # (a bit duplicate, no?)
+        sys.exit(1)
 
 
 if __name__ == '__main__':
