@@ -1,10 +1,12 @@
 import os
 import re
 import subprocess
+import time
 
 from . import settings
-from .nvmedata import NvmeDataParser
-from .smartdata import SmartDataParser
+from .devtype import DEVTYPE_HDD, DEVTYPE_NVME, DEVTYPE_SSD
+from .nvmedata import NvmeDataError, NvmeDataParser
+from .smartdata import SmartDataError, SmartDataParser
 from .hdparm import HdparmParser
 
 __author__ = 'osso'
@@ -15,13 +17,13 @@ class UnmatchedStorageDevice(Exception):
 
 
 class BaseStorageDevice(object):
-    DEVTYPE_HDD = 'hdd'     # plain old spinner
-    DEVTYPE_NVME = 'nvme'   # NVMe SSD
-    DEVTYPE_SSD = 'ssd'     # Solid State Drive (SSD)
-
     @classmethod
     def from_devname(cls, devname):
-        for class_ in (NvmeStorageDevice, GenericStorageDevice):
+        for class_ in (
+                NvmeStorageDevice,
+                SgSanitizeStorageDevice,
+                FallbackStorageDevice,
+                ):
             try:
                 devtype, hwdata = class_._get_hwdata(devname)
             except UnmatchedStorageDevice as e:
@@ -70,7 +72,7 @@ class BaseStorageDevice(object):
         return self.hwdata.ata_version
 
     def get_smart_status(self):
-        # "PASSED" or "FAILED!" or "Enabled" or "MISSING"
+        # "PASSED" or "OK" or "FAILED!" or "Enabled" or "MISSING"
         return self.hwdata.smart_status
 
     def is_sas(self):
@@ -78,7 +80,7 @@ class BaseStorageDevice(object):
             'Transport protocol', '').lower()
 
     def is_ssd(self):
-        return (self.devtype in (self.DEVTYPE_NVME, self.DEVTYPE_SSD))
+        return (self.devtype in (DEVTYPE_NVME, DEVTYPE_SSD))
 
     def _get_docktool_bay_nr(self, port):
         return settings.ATA_TO_BAY_NR.get(port, None)
@@ -172,134 +174,257 @@ class BaseStorageDevice(object):
             subprocess.check_output([
                 '/usr/bin/sdparm', '--readonly', '--command=stop',
                 self.devname])
-        elif self.devtype == self.DEVTYPE_NVME:
+        elif self.devtype == DEVTYPE_NVME:
             print('WARNING: No idea how to eject and NVMe dev; needed?')
         else:
-            assert self.devtype in (self.DEVTYPE_HDD, self.DEVTYPE_SSD)
+            assert self.devtype in (DEVTYPE_HDD, DEVTYPE_SSD)
             # FIXME: what if we don't want to spin it down? because we're not
             # ejecting it
             subprocess.check_output(['/sbin/hdparm', '-Y', self.devname])
         self.status = 'Shutdown'
 
 
-class GenericStorageDevice(BaseStorageDevice):
+class GenericStorageDeviceBase(BaseStorageDevice):
     @classmethod
     def _get_hwdata(cls, devname):
         try:
             hwdata = SmartDataParser(devname)
-        except Exception as e:
+        except SmartDataError as e:
             raise UnmatchedStorageDevice(e) from e
+        if hwdata.devtype == DEVTYPE_NVME:
+            raise NotImplementedError(
+                'programming error; we should be in NvmeStorageDevice')
+        return hwdata.devtype, hwdata
 
-        if 'NVMe' in hwdata.rawdata:
-            devtype = cls.DEVTYPE_NVME
-            assert False, 'we should be in NvmeStorageDevice'
-        elif 'solid state' in hwdata.information.get(
-                'Rotation Rate', '').lower():
-            devtype = cls.DEVTYPE_SSD
+
+class SgSanitizeStorageDevice(GenericStorageDeviceBase):
+    @classmethod
+    def _get_hwdata(cls, devname):
+        devtype, hwdata = super()._get_hwdata(devname)
+
+        # Is this as SAS disk, then assume hdparm security yields None
+        # and we might get crypto erase support.
+        if 'sas' in hwdata.information.get('Transport protocol', '').lower():
+            # Assert that a SAS disk does not do hdparm stuff.
+            hdparm_sec_info = HdparmParser(devname).get_security_info()
+            assert hdparm_sec_info is None, (
+                hdparm_sec_info, hwdata.information)
+
+            # Try sg_opcodes.
+            opcodes = cls._get_sg_security_info(devname)
+            if opcodes['sanitize_crypto_erase']:
+                return devtype, hwdata
+
+        raise UnmatchedStorageDevice()
+
+    def get_erase_methods(self):
+        return [
+            self.wipe_using_sanitize_crypto_erase,
+            self.wipe_using_manual_methods]
+
+    def wipe_using_sanitize_crypto_erase(self):
+        # /usr/bin/sg_sanitize --quick --wait --verbose --crypto devname
+        if not settings.DEBUG:
+            if not (
+                    self._get_sg_security_info(self.devname)
+                    ['sanitize_crypto_erase']):
+                return False, 'no sanitize_crypto_erase available'
+            subprocess.check_call([
+                '/usr/bin/sg_sanitize', '--quick', '--wait', '--verbose',
+                '--crypto', self.devname])
+            return True, None
         else:
-            devtype = cls.DEVTYPE_HDD
+            print('WARNING: Skipping crypto erase erase because DEBUG')
+            return True, None
+    wipe_using_sanitize_crypto_erase.name = 'sanitize crypto erase'
+    wipe_using_sanitize_crypto_erase.type = 'builtin'  # fast!
 
-        return devtype, hwdata
+    @classmethod
+    def _get_sg_security_info(cls, devname):
+        opcodes = cls._get_sg_opcodes(devname)
+        ret = {
+            # 'sanitize_overwrite': False,
+            'sanitize_crypto_erase': False,
+        }
+        if (0x48, '1') in opcodes:
+            # This, is no fun. I tested one Seagate 10TB disk, and it did
+            # 175MB/s for the first 3.78% (in 36 minutes). That translates to
+            # 16 hours for the total disk.
+            # # sg_requests /dev/sda
+            # > Decode parameter data as sense data:
+            # > Fixed format, current; Sense key: Not Ready
+            # > Additional sense: Logical unit not ready, sanitize in progress
+            # >   Progress indication: 3.78%
+            # Additionally, it might even have stalled now, as the status
+            # update from sg_requests totally hung at t0+4hours.
+            # #ret['sanitize_overwrite'] = True
+            pass
+        if (0x48, '3') in opcodes:
+            ret['sanitize_crypto_erase'] = True
+        return ret
 
+    @classmethod
+    def _get_sg_opcodes(cls, devname):
+        # > # sg_opcodes --compact /dev/sda
+        # >   SEAGATE   ST10000NM0226     KT01
+        # >   Peripheral device type: disk
+        # >
+        # > Opcode,sa  Name
+        # >   (hex)
+        # > ---------------------------------------
+        # >  00        Test Unit Ready
+        # >  01        Rezero Unit
+        # >  03        Request Sense
+        # > ...
+        # >  48,1      Sanitize, overwrite
+        # >  48,3      Sanitize, cryptographic erase
+        # >  48,1f     Sanitize, exit failure mode
+        opcodes = (
+            subprocess.check_output(
+                ['/usr/bin/sg_opcodes', '--compact', devname])
+            .decode('utf-8'))
+        it = iter(opcodes.strip().split('\n'))
+        for line in it:
+            if line.startswith('----'):
+                break
+        ret = {}
+        for line in it:
+            opcode, title = line.strip().split(None, 1)
+            opcodes = opcode.split(',')
+            opcodes[0] = int(opcodes[0], 16)
+            assert tuple(opcodes) not in ret, (ret, opcodes)
+            ret[tuple(opcodes)] = title
+        return ret
+
+
+class FallbackStorageDevice(GenericStorageDeviceBase):
     def get_erase_methods(self):
         if self._can_secure_erase()[0]:
             return [
                 self.wipe_using_security_erase, self.wipe_using_manual_methods]
         return [self.wipe_using_manual_methods]
 
-    def wipe_using_security_erase(self):
-        # s1 = /dev/sdc
-        # secret=`pwgen -s 32 1`
-        # hdparm --user-master u --security-set-pass $secret $1
-        # hdparm --user-master u --security-erase $secret $1
-
-        if not settings.DEBUG:
-            can_erase, error = self._can_secure_erase()
-            if not can_erase:
-                return False, error
-
-            # Create secret
-            # XXX: this is problematic if the system dies before we've called
-            # security-erase: afterwards the disk sometimes stays locked. It'd
-            # be nicer if we wiped with a pre-existing pw?
-            secret = subprocess.check_output([
-                '/usr/bin/pwgen', '-s', '32', '1'])[0:32].decode('utf-8')
-
-            # Set the password. This should set 'enabled' in hdparm
-            # security info
-            subprocess.check_output([
-                '/sbin/hdparm', '--user-master', 'u',
-                '--security-set-pass', secret, self.devname])
-
-            # Check 'enabled' in hdparm security info
-            security_info = self._hdparm_security_info()
-            if not security_info.get('enabled', False):
-                return False, (
-                    'Block device password could not be set, '
-                    'security info: {}'.format(security_info))
-
-            # Only perform secure BLKDEV erase if 'enabled'
-            subprocess.check_output([
-                '/sbin/hdparm', '--user-master', 'u',
-                '--security-erase', secret, self.devname])
-
-            # Check that 'enabled' is again False
-            security_info = self._hdparm_security_info()
-
-            if security_info.get('enabled', True):
-                return False, (
-                    'Block device password could not be UNset, '
-                    'security info: {}, secret: {}'.format(
-                        security_info, secret))
-            secret = None
-        else:
-            print('WARNING: Skipping secure erase because DEBUG')
-
-        return True, None
-    wipe_using_security_erase.name = 'secure SSD/HDD erase'
-    wipe_using_security_erase.type = 'builtin'  # slow? fast?
-
     def _can_secure_erase(self):
-        security_info = self._hdparm_security_info()
+        security_info = HdparmParser(self.devname).get_security_info()
+        if security_info is None:
+            return False, 'no hdparm security, no sg_opcode security'
 
-        for x in ['supported', 'enabled', 'locked', 'frozen']:
+        for x in ['has_security', 'is_enabled', 'is_locked', 'is_frozen']:
             if x not in security_info:
                 return False, '{} not found in security_info: {}'.format(
                     x, security_info)
 
-        if not security_info['supported']:
+        if not security_info['has_security']:
             return False, 'Block device secure erase seems not to be supported'
 
-        if security_info['enabled']:
-            return False, (
+        if security_info['is_enabled']:
+            assert False, (
                 'Block device password for secure erase is already set')
 
-        if security_info['locked']:
-            return False, (
+        if security_info['is_locked']:
+            assert False, (
                 'Block device is locked (try replugging the disk or '
                 'hibernating system)')
 
-        if security_info['frozen']:
-            return False, (
+        if security_info['is_frozen']:
+            assert False, (
                 'Block device is frozen (try replugging the disk or '
                 'hibernating system)')
 
+        if security_info.get('is_past_retry_count'):
+            assert False, (
+                'Block device is past password retry count (time to look up '
+                'a master password?)')
+
         return True, None
 
-    def _hdparm_security_info(self):
-        hdparm_parser = HdparmParser(self.devname)
-        return hdparm_parser.get_security_info()
+    def wipe_using_security_erase(self):
+        # $1 = /dev/sdc
+        # secret=`pwgen -s 32 1`
+        # hdparm --user-master u --security-set-pass $secret $1
+        # hdparm --user-master u --security-erase $secret $1
+        if not settings.DEBUG:
+            can_erase, error = self._can_secure_erase()
+            if not can_erase:
+                return False, error
+            return self._wipe_using_hdparm_security_erase()
+        else:
+            print('WARNING: Skipping secure erase because DEBUG')
+            return True, None
+    wipe_using_security_erase.name = 'secure SSD/HDD erase'
+    wipe_using_security_erase.type = 'builtin'  # slow? fast?
+
+    def _wipe_using_hdparm_security_erase(self):
+        # Create secret
+        # XXX: this is problematic if the system dies before we've called
+        # security-erase: afterwards the disk sometimes stays locked. It'd
+        # be nicer if we wiped with a pre-existing pw?
+        secret = subprocess.check_output([
+            '/usr/bin/pwgen', '-s', '32', '1'])[0:32].decode('utf-8')
+
+        # Set the password. This should set is_enabled ('enabled')
+        # in hdparm security info.
+        print('WARNING: The hdparm --security-set-pass sometimes says: ')
+        print('  "SG_IO: bad/missing sense data, sb[]:  70 00 01 00 ..."')
+        print('This might be ignorable. Please inform developer.')
+        print(
+            'NOTICE: Using secret {!r}.. you may need this to unbrick!'
+            .format(secret))
+
+        print('> --security-set-pass')
+        subprocess.check_output([
+            '/sbin/hdparm', '--user-master', 'u',
+            '--security-set-pass', secret, self.devname])
+
+        # Check hdparm security info. Only perform secure BLKDEV
+        # erase if security is indeed enabedl.
+        security_info = HdparmParser(self.devname).get_security_info()
+        print('> {}'.format(security_info))
+        if not security_info.get('is_enabled'):
+            return False, (
+                'Block device password could not be set, '
+                'security info: {}'.format(security_info))
+
+        for i in range(10):
+            # Try secure erase.
+            if security_info.get('has_enhanced_erase', False):
+                print('> --security-erase-enhanced')
+                subprocess.check_output([
+                    '/sbin/hdparm', '--user-master', 'u',
+                    '--security-erase-enhanced', secret, self.devname])
+            else:
+                print('> --security-erase')
+                subprocess.check_output([
+                    '/sbin/hdparm', '--user-master', 'u',
+                    '--security-erase', secret, self.devname])
+            # Check that hdparm security is now unset.
+            security_info = HdparmParser(self.devname).get_security_info()
+            print('> {}'.format(security_info))
+            if not security_info.get('is_enabled', True):
+                break
+            # This could be a problem.. retry this a few times..
+            print(
+                'WARNING: Security erase/unset failed (password {!r});'
+                'retrying in 3..'.format(secret))
+            time.sleep(3)
+        else:
+            return False, (
+                'Block device password could not be UNset, '
+                'security info: {}, secret: {!r}'.format(
+                    security_info, secret))
+
+        return True, None
 
 
 class NvmeStorageDevice(BaseStorageDevice):
     @classmethod
     def _get_hwdata(cls, devname):
-        devtype = cls.DEVTYPE_NVME
         try:
             hwdata = NvmeDataParser(devname)
-        except Exception as e:
+        except NvmeDataError as e:
             raise UnmatchedStorageDevice(e) from e
-        return devtype, hwdata
+        assert hwdata.devtype == DEVTYPE_NVME, hwdata.devtype
+        return hwdata.devtype, hwdata
 
     def get_erase_methods(self):
         methods = []
@@ -368,7 +493,7 @@ class NvmeStorageDevice(BaseStorageDevice):
     wipe_using_crypto_sanitize.type = 'builtin'  # ???
 
     def _get_secure_erase_options(self):
-        if self.devtype != self.DEVTYPE_NVME:
+        if self.devtype != DEVTYPE_NVME:
             return False, 'Disk was not recognized as NVMe Device'
 
         # > Secure erase is faster, but not as complete. Data on a drive that
