@@ -5,14 +5,24 @@ import os.path
 import re
 import subprocess
 
+from .devtype import DEVTYPE_NVME
+
 __author__ = 'osso'
 
 
+def _assert_int(value):
+    "Precondition or postcondition to check that this is an int"
+    assert isinstance(value, int), value
+    return value
+
+
 class NvmeDataError(Exception):
     pass
 
 
 class NvmeDataParser(object):
+    devtype = DEVTYPE_NVME
+
     def __init__(self, devname):
         self.devname = devname
         # Yes, nvme cli has --output-format=json too, but it lacks explanation
@@ -30,7 +40,7 @@ class NvmeDataParser(object):
         nvmecli = '/usr/sbin/nvme'
         try:
             p = subprocess.Popen(
-                [nvmecli] + cmd + [self.devname],
+                [nvmecli] + cmd + [self.devname], env={'LC_ALL': 'C'},
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
             stdout, stderr = p.communicate()
         except FileNotFoundError as e:
@@ -149,15 +159,15 @@ class NvmeDataParser(object):
 
     @property
     def power_on_hours(self):
-        return int(self.information['power_on_hours'].replace(',', ''))
+        return _assert_int(self.information['power_on_hours'])
 
     @property
     def lbas_read(self):
-        return int(self.information['host_read_commands'].replace(',', ''))
+        return _assert_int(self.information['host_read_commands'])
 
     @property
     def lbas_written(self):
-        return int(self.information['host_write_commands'].replace(',', ''))
+        return _assert_int(self.information['host_write_commands'])
 
     @property
     def reallocated_sector_ct(self):
@@ -182,7 +192,9 @@ class NvmeDataParser(object):
         #   [4:4] : 0   Firmware Activate Without Reset Not Supported
         #   [3:1] : 0x1	Number of Firmware Slots
         #   [0:0] : 0   Firmware Slot 1 Read/Write
-        bitre = re.compile(r'\s*\[(\d+):(\d+)\]\s*:\s*([0-9x]+)\s+(.*)$')
+        # ...
+        #   [31:1] : 0x6a\tReserved
+        bitre = re.compile(r'\s*\[(\d+):(\d+)\]\s*:\s*([0-9a-fx]+)\s+(.*)$')
 
         information = {}
         last_key = None
@@ -198,28 +210,12 @@ class NvmeDataParser(object):
                 key = key.strip()
 
                 if key.startswith('['):
-                    assert last_key  # must be something
                     # [31:30] : 0x2 Blah <-- bits 30+31 have 31 set
-                    m = bitre.match(line)
-                    assert m, line
-                    bitend, bitstart, value, name = m.groups()
-                    if bitend == bitstart:
-                        value = bool(int(value, 16))    # 0, 0x1
-                        information['{}:{}'.format(last_key, bitend)] = value
-                    else:
-                        value = int(value, 16)          # 0, 0x1, 0x2, ...
-                        information['{}:{}:{}'.format(
-                            last_key, bitend, bitstart)] = value
-
+                    self._populate_subinfo(
+                        information, key, value, last_key, line, bitre)
                 elif not indent:
                     last_key = key
-                    value = value.strip()
-                    if value.startswith('0x') or value == '0':
-                        value = int(value, 16)
-                    if key not in information:
-                        information[key] = value
-                    else:
-                        assert False, ((key, value), information)
+                    self._populate_info(information, key, value)
 
         attributes = {}
         if not attributes:
@@ -227,6 +223,36 @@ class NvmeDataParser(object):
 
         return information, attributes
 
+    def _populate_info(self, information, key, value):
+        # "frmw      : 0x2" => {"frmw": 0x02}
+        value = value.strip()
+        if value.startswith('0x') or value == '0':
+            value = int(value, 16)  # 0x12 -> 18
+        elif all(group.isdigit() for group in value.split(',')) and (
+                all(len(group) == 3 for group in value.split(',')[1:])):
+            value = int(value.replace(',', ''))  # 1,234,567 -> 1234567
+        if key not in information:
+            information[key] = value
+        else:
+            assert False, ((key, value), information)
+
+    def _populate_subinfo(self, information, key, value, last_key, line,
+                          bitre):
+        assert last_key  # must be something
+        # > [31:30] : 0x2 Blah <-- bits 30+31 have 31 set
+        # " [31:30] : 0x2 Blah" => {"frmw:30:31": 0x2}
+        m = bitre.match(line)
+        assert m, (
+            'does not match bitre: {}'.format(bitre.pattern), line)
+        bitend, bitstart, value, name = m.groups()
+        if bitend == bitstart:
+            value = bool(int(value, 16))    # 0, 0x1
+            information['{}:{}'.format(last_key, bitend)] = value
+        else:
+            value = int(value, 16)          # 0, 0x1, 0x2, ...
+            information['{}:{}:{}'.format(
+                last_key, bitend, bitstart)] = value
+
 
 if __name__ == "__main__":
     import sys
