@@ -19,13 +19,16 @@ ATA_TO_BAY_NR = {
     'ata-3': 4,
 }
 
+# List of devices that are ignored for udev/auto notifications.
+IGNORE_DEVICES = []
+
 # disk empty sample settings
 # - divide disk in sample_count regions
 # - sample sample_size bytes randomly per region
-POST_WIPE_SAMPLE_COUNT = 100
-POST_WIPE_SAMPLE_SIZE = 'MB'
+POST_WIPE_SAMPLE_COUNT = 1000
+POST_WIPE_SAMPLE_SIZE = 100 * 1024
 
-# Find local_settings.py in /etc/osso-docktool or next to this file.
+# Find local_settings.py in /etc/osso-docktool or in the parent directory.
 sys.path.insert(0, '/etc/osso-docktool')
 try:
     from local_settings import *  # noqa
